extern crate base64;

use core::cmp::Ordering;
use std::fs::{File, metadata};
use std::io::{BufReader, BufWriter, Cursor, Error, Read, Seek, SeekFrom, Write};
use std::ops::{Shl, Shr};
use std::time::SystemTime;

use rug::{Integer, integer::IsPrime, integer::Order, rand::RandState};

use clap::{App, Arg};


const BLOCK_SIZE: usize = 8;
const ETA: u32 = 511;
const RHO: u32 = 160;


// region Utils

fn init_random_state() -> RandState<'static> {
    let mut random_state: RandState = RandState::new();
    let seed = SystemTime::now().duration_since(SystemTime::UNIX_EPOCH)
        .expect("Duration since UNIX_EPOCH failed");
    random_state.seed(&Integer::from(seed.as_secs()));
    return random_state
}

fn gen_random_integer(random_state: &mut RandState, lbound: &Integer, rbound: &Integer) -> Integer {
    let diff = rbound - lbound;
    let i = Integer::from(diff);
    i.random_below(random_state) + lbound
}

fn gen_prime_integer(random_state: &mut RandState, lbound: &Integer, rbound: &Integer) -> Integer {
    return loop {
        let mut p = gen_random_integer(random_state, &lbound, &rbound);
        p.set_bit(0, true);
        match p.is_probably_prime(50) {
            IsPrime::No => continue,
            IsPrime::Probably | IsPrime::Yes => break p
        }
    };
}

fn crt_init(ps: &Vec<Integer>) -> Vec<Integer> {
    let m = ps.iter().fold(Integer::from(1), |prod, val| prod * val);
    let mdps: Vec<Integer> = ps.iter().map(|p| Integer::from(&m) / p).collect();
    let mdps_inv: Vec<Integer> = mdps.iter().zip(ps)
        .map(|v| Integer::from(v.0).invert(v.1).unwrap()).collect();
    let c = mdps.iter().zip(mdps_inv)
        .map(|v| (v.0 * v.1) % Integer::from(&m)).collect();
    return c
}

fn crt_final(a: &Vec<Integer>, c: &Vec<Integer>, m: &Integer) -> Integer {
    let prod: Vec<Integer> = a.iter().zip(c)
        .map(|v| Integer::from(v.0 * v.1) % Integer::from(m)).collect();
    let v = prod.iter().fold(Integer::from(0), |sum, val| sum + val) % m;
    return if v > Integer::from(m).shr(1) { v - m } else { v }
}


fn write_integer<W: Write>(a: &Integer, f: &mut BufWriter<W>) -> Result<(), Error> {
    let sign = if a.cmp0() == Ordering::Less { [0xffu8] } else { [0x01u8] };
    let digits = a.to_digits::<u8>(Order::LsfLe);
    f.write_all(&(digits.len() as u32).to_le_bytes())?;
    f.write_all(&sign)?;
    f.write_all(&digits)?;
    Ok(())
}

fn read_integer<R: Read>(f: &mut BufReader<R>) -> Result<Integer, Error> {
    let mut length = [0u8; 4];
    f.read_exact(&mut length)?;
    let length = u32::from_le_bytes(length);

    let mut sign = [0u8; 4];
    f.read_exact(&mut sign[..1])?;
    let sign = u32::from_le_bytes(sign);

    let mut digits: Vec<u8> = vec![0u8; length as usize];
    f.read_exact(&mut digits)?;
    let a = Integer::from_digits(&digits, Order::LsfLe);

    return Ok(if sign == 0xff { Integer::from(&*a.as_neg()) } else { a })
}

// endregion Utils


// region Key-related functions

fn gen_key_file<W: Write>(f: &mut BufWriter<W>, eta: u32, rho: u32) -> Result<(Vec<Integer>, Integer), Error> {
    // 1. initialize the RNG
    let mut random_state = init_random_state();
    let mut lbound = Integer::from(0); lbound.set_bit(eta - 1, true);
    let mut rbound = Integer::from(0); rbound.set_bit(eta, true);

    // 2. generate the primes
    let ps: Vec<Integer> = (0..BLOCK_SIZE).map(|_| gen_prime_integer(&mut random_state, &lbound, &rbound)).collect();
    let m = ps.iter().fold(Integer::from(1), |prod, val| prod * val);
    assert!(rho < eta);

    // 3. write serialized integers to a buffer
    let mut fout = BufWriter::new(Cursor::new(Vec::new()));
    ps.iter().for_each(|p| write_integer(p, &mut fout).unwrap());
    write_integer(&Integer::from(eta), &mut fout)?;
    write_integer(&Integer::from(rho), &mut fout)?;

    // 4. b64-encode the data and write to file
    let mut buf: Cursor<Vec<u8>> = fout.into_inner().unwrap();
    buf.seek(SeekFrom::Start(0)).unwrap();
    let data: Vec<u8> = buf.into_inner();
    f.write_all("-----BEGIN KEY-----\n".as_ref())?;
    f.write_all(base64::encode(&data).as_ref())?;
    f.write_all("\n------END KEY------".as_ref())?;

    return Ok((ps, m))
}

fn read_key_file<R: Read>(f: &mut BufReader<R>) -> Result<(Vec<Integer>, Integer, u32, u32), Error> {
    // 1. skip the header and read the raw bytes
    let mut data: Vec<u8> = Vec::new();
    f.read_to_end(&mut data)?;
    data.resize(data.len() - 20, 0);
    data.drain(0..20);

    // 2. 64-decode the data
    data = base64::decode(&data).unwrap();
    let mut fin = BufReader::new(Cursor::new(data));

    // 3. read the primes and the parameters
    let ps: Vec<Integer> = (0..BLOCK_SIZE).map(|_| read_integer(&mut fin).unwrap()).collect();
    let eta = read_integer(&mut fin)?;
    let rho = read_integer(&mut fin)?;

    let m = ps.iter().fold(Integer::from(1), |prod, val| prod * val);
    return Ok((ps, m, eta.to_u32().unwrap(), rho.to_u32().unwrap()))
}

// endregion Key-related functions


// region The Cipher

fn encrypt_block(random_state: &mut RandState, block: &Vec<u8>, m: &Integer, rho: u32, crtc: &Vec<Integer>) -> Integer {
    assert_eq!(block.len(), BLOCK_SIZE);
    let mut rbound = Integer::from(0); rbound.set_bit(rho - 8, true);
    let mut lbound = Integer::from(0); lbound.set_bit(rho - 8, true);
    lbound = -lbound + 1;
    let rs: Vec<Integer> = block.iter()
        .map(|v| gen_random_integer(random_state, &lbound, &rbound).shl(8) + Integer::from(*v)).collect();
    let c = crt_final(&rs, &crtc, &m);
    return c
}

fn decrypt_block(c: &Integer, ps: &Vec<Integer>) -> Vec<u8> {
    return ps.iter().map(|p| {
        let mut v = Integer::from(c) % p;
        v = if v < 0 { p + v } else { v };
        v = if v > Integer::from(p).shr(1) { v - p } else { v };
        v = v % 256;
        v = if v < 0 { 256 + v } else { v };
        v.to_u8_wrapping()
    }).collect::<Vec<u8>>()
}

fn encrypt<R: Read, W: Write>(fin: &mut BufReader<R>, fout: &mut BufWriter<W>, size: usize, ps: &Vec<Integer>, m: &Integer, rho: u32) -> Result<(), Error> {
    // 1. prepare data for CRT calculation and init the RNG
    let mut random_state = init_random_state();
    let crtc = crt_init(ps);

    // 2. encrypt the data block-by-block and write to the output IO buffer
    let mut mb = vec![0u8; BLOCK_SIZE];
    for _ in 0..(size / BLOCK_SIZE) {
        fin.read_exact(&mut mb)?;
        let c = encrypt_block(&mut random_state, &mb, m, rho, &crtc);
        write_integer(&c, fout)?;
    }

    // 3. try to read the last block and pad it prior to encryption
    let n_rem = size % BLOCK_SIZE;
    fin.read_exact(&mut mb[..n_rem])?;
    let pos = if n_rem == BLOCK_SIZE { 0 } else { n_rem };
    mb[pos] = 0x80;
    for i in (pos + 1)..BLOCK_SIZE {
        mb[i] = 0x00;
    }
    let c = encrypt_block(&mut random_state, &mb, m, rho, &crtc);
    write_integer(&c, fout)?;
    Ok(())
}

fn decrypt<R: Read, W: Write>(fin: &mut BufReader<R>, fout: &mut BufWriter<W>, ps: &Vec<Integer>) -> Result<(), Error> {
    let mut mb: Vec<u8>;
    let mut mb_prev: Vec<u8>;

    // 1. decrypt the first block and keep for now
    let c = match read_integer(fin) {
        Ok(c) => c,
        Err(_) => return Ok(())
    };
    mb_prev = decrypt_block(&c, ps);

    // 2. decrypt the next blocks and write to fout
    loop {
        let c = match read_integer(fin) {
            Ok(c) => c,
            Err(_) => break
        };
        mb = decrypt_block(&c, ps);
        fout.write_all(&mb_prev)?;
        mb_prev = mb;
    }

    // 3. at this point all the blocks have been decrypted and the last one (mb_prev) is to be written
    //    so we strip the padding and write what's left of it to the fout
    let mut pos = 0;
    while pos < BLOCK_SIZE {
        match mb_prev[BLOCK_SIZE - 1 - pos] {
            0x00 => { pos += 1; },
            0x80 => { pos += 1; break },
            _ => break  // TODO: panic here?
        }
    }
    if pos < BLOCK_SIZE {
        mb_prev.resize(BLOCK_SIZE - pos, 0);
        fout.write_all(&mb_prev)?;
    }

    Ok(())
}

// endregion The Cipher


fn main() -> Result<(), Error> {
    let matches = App::new("Regular Cipher")
        .version("1.0")
        .about("Encrypts your data with an unbreakable encryption algorithm!")
        .subcommand(App::new("genkey")
            .about("generate encryption key file")
            .arg(Arg::new("PATH")
                .about("Path to output key file")
                .required(true)
                .index(1)))
        .subcommand(App::new("encrypt")
            .about("encrypt data file")
            .arg(Arg::new("INPUT")
                .about("Path to input file")
                .required(true)
                .index(1))
            .arg(Arg::new("OUTPUT")
                .about("Path to output file")
                .required(true)
                .index(2))
            .arg(Arg::new("KEY")
                .about("Path to key file")
                .required(true)
                .index(3)))
        .subcommand(App::new("decrypt")
            .about("decrypt data file")
            .arg(Arg::new("INPUT")
                .about("Path to input file")
                .required(true)
                .index(1))
            .arg(Arg::new("OUTPUT")
                .about("Path to output file")
                .required(true)
                .index(2))
            .arg(Arg::new("KEY")
                .about("Path to key file")
                .required(true)
                .index(3)))
        .get_matches();

    match matches.subcommand() {
        Some(("genkey", genkey_matches)) => {
            let f = File::create(genkey_matches.value_of("PATH").unwrap())?;
            let mut fout = BufWriter::new(f);
            gen_key_file(&mut fout, ETA, RHO)?;
        }
        Some(("encrypt", enc_matches)) => {
            let path_in = enc_matches.value_of("INPUT").unwrap();
            let fin = File::open(path_in)?;
            let fout = File::create(enc_matches.value_of("OUTPUT").unwrap())?;
            let fkey = File::open(enc_matches.value_of("KEY").unwrap())?;
            let mut fin = BufReader::new(fin);
            let mut fout = BufWriter::new(fout);
            let mut fkey = BufReader::new(fkey);

            let (ps, m, _, _) = read_key_file(&mut fkey)?;
            let meta = metadata(path_in)?;
            encrypt(&mut fin, &mut fout, meta.len() as usize, &ps, &m, RHO)?;
        }
        Some(("decrypt", dec_matches)) => {
            let fin = File::open(dec_matches.value_of("INPUT").unwrap())?;
            let fout = File::create(dec_matches.value_of("OUTPUT").unwrap())?;
            let fkey = File::open(dec_matches.value_of("KEY").unwrap())?;
            let mut fin = BufReader::new(fin);
            let mut fout = BufWriter::new(fout);
            let mut fkey = BufReader::new(fkey);

            let (ps, _, _, _) = read_key_file(&mut fkey)?;
            decrypt(&mut fin, &mut fout, &ps)?;
        }
        None => println!("No subcommand was used"),
        _ => unreachable!(),
    }

    Ok(())
}
